import numpy as np
import matplotlib.pyplot as plt
import matplotlib
from math import exp
import itertools
import math
import EM_Model
import Unsupervised_Model
import Two_Stage_MLE_Model

ground_truth = 10

# generate simulation data
num_workers = 1000
num_data = 10000
worker_mu = np.random.normal(0, 1, num_workers)
worker_var = np.random.lognormal(0, 1, num_workers)
worker_data = np.zeros((num_workers,num_data))
for i in range(num_workers):
    worker_data[i,:] = np.random.normal(ground_truth+worker_mu[i],worker_var[i],num_data)

worker_ids = [i for i in range(num_workers)]
# EM model
em_model = EM_Model(worker_data)
em_model.train_worker_set(np.array(worker_ids))

# unsupervised model
unsupervised_model = Unsupervised_Model(worker_data)
unsupervised_model.train_worker_set(np.array(worker_ids))

# two stage mle model
data_label = np.full((num_data,1), ground_truth)
two_stage_mle_model = Two_Stage_MLE_Model(worker_data,data_label)
two_stage_mle_model.train_worker_set(np.array(worker_ids))

# compute linear reward
norm_reputation_em = em_model.reputation_worker_em/np.sum(em_model.reputation_worker_em)
norm_reputation_unsupervised = unsupervised_model.reputation_worker_unsupervised/np.sum(unsupervised_model.reputation_worker_unsupervised)
norm_reputation_mle = two_stage_mle_model.reputation_mle/np.sum(two_stage_mle_model.reputation_mle)
true_worker_var = worker_var**2
true_reputation = 1/true_worker_var
norm_true_reputation = true_reputation/np.sum(true_reputation)

# compute Shapley value reward
# EM model
marginal_contribution_em = np.zeros(len(worker_ids))
for worker_id_perm in itertools.permutations(worker_ids):
    contribution_acc = 0
    for i in range(1,len(worker_ids)+1):
        em_model.train_worker_set(np.array(worker_id_perm[:i]))
        marginal_contribution_em[worker_id_perm[i-1]-1] = marginal_contribution_em[worker_id_perm[i-1]-1] + np.sum(em_model.reputation_worker_em) - contribution_acc
        contribution_acc = np.sum(em_model.reputation_worker_em)
        
# Unsupervised model
marginal_contribution_unsupervised = np.zeros(len(worker_ids))
for worker_id_perm in itertools.permutations(worker_ids):
    contribution_acc = 0
    for i in range(1,len(worker_ids)+1):
        unsupervised_model.train_worker_set(np.array(worker_id_perm[:i]))
        marginal_contribution_unsupervised[worker_id_perm[i-1]-1] = marginal_contribution_unsupervised[worker_id_perm[i-1]-1] + np.sum(em_model.reputation_worker_em) - contribution_acc
        contribution_acc = np.sum(unsupervised_model.reputation_worker_unsupervised)

# MLE
marginal_contribution_mle = np.zeros(len(worker_ids))
for worker_id_perm in itertools.permutations(worker_ids):
    contribution_acc = 0
    for i in range(1,len(worker_ids)+1):
        two_stage_mle_model.train_worker_set(np.array(worker_id_perm[:i]))
        marginal_contribution_mle[worker_id_perm[i-1]-1] = marginal_contribution_mle[worker_id_perm[i-1]-1] + np.sum(em_model.reputation_worker_em) - contribution_acc
        contribution_acc = np.sum(two_stage_mle_model.reputation_mle)
